/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Doug Daniels.
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Matthew Wilson
 *    Dr. Katsuhiko Tsujino
 *    Eliot Setzer
 *    C. Greg Hagerty
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

// TODO combine with AnnozillaAnnotationModel?
function AnnozillaIconDataModel (target)
{
    this.target = target;
}

AnnozillaIconDataModel.ANNOZILLA_ICONS_KEY = "annozilla-icons";

AnnozillaIconDataModel.prototype.getIcons =
function ()
{
    return this.target.getUserData (AnnozillaIconDataModel.ANNOZILLA_ICONS_KEY);
}

AnnozillaIconDataModel.prototype.setIcons =
function (icons)
{
    this.target.setUserData
       (AnnozillaIconDataModel.ANNOZILLA_ICONS_KEY,
        icons,
        function (operation, key, data, source, destination) { } );
}

AnnozillaIconDataModel.prototype.clearIcons = function ()
{
    this.setIcons ( new Object() );
}

function AnnozillaIconManager(annozilla0) {

    if (! AnnozillaIconManager._hasRequiredComponents()) {
        return;
    }

    this.annozilla = annozilla0;

    // whether we're inserting icons
    this.insertIcons = false;
    // whether we're inserting highlighting
    this.insertHighlight = false;

    // say that we are not observing annotations -- this must be done before loading prefs
    this.observing = false;

    // bundle
    this.bundle = srGetStrBundle("chrome://annozilla/locale/annozilla.properties");

    var that = this;
    // add a listener for when the window closes
    window.addEventListener("unload", function(e) { that.unload(); }, false);

    // TODO multiple tabs
    new AnnozillaIconDataModel (gBrowser.selectedBrowser).clearIcons();

}

AnnozillaIconManager.prototype = new AnnozillaAnnotationsView();

//also used by ContextMenu.js
AnnozillaIconManager.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
AnnozillaIconManager.DC10_NS  = "http://purl.org/dc/elements/1.0/";
AnnozillaIconManager.TYPE_NS  = "http://www.w3.org/2000/10/annotationType#";
AnnozillaIconManager.ANNOTATION_ID_ATTRIBUTE_NAME = "annotation-id";
AnnozillaIconManager.ANNOTATION_TYPE_ATTRIBUTE_NAME = "annotation-type";

/*******************
 * PUBLIC FUNCTIONS
 *******************/

AnnozillaIconManager.prototype.QueryInterface = function(iid) {  
  if (! (iid.equals(Components.interfaces.nsIAnnotationsObserver) ||
         iid.equals(Components.interfaces.nsISupports) ) ) {
    throw Components.results.NS_ERROR_NO_INTERFACE;
  }
  return this;
}

AnnozillaIconManager.prototype.unload = function() {
    if(this.observing){
        // remove observer
        this.annozilla.app.removeObserver(this);
        this.observing = false;
    }
}
  
// called whenever the insertIcons pref is set
AnnozillaIconManager.prototype.insertIconsPrefChanging = function(insert) {
    this.insertIcons = insert;
    this._insertPrefChange();
}

// called whenever the insertHighlight pref is set
AnnozillaIconManager.prototype.insertHighlightPrefChanging = function(insert) {
    this.insertHighlight = insert;
    this._insertPrefChange();
}

/**
 * Called when an annotation load is beginning.
 * @param target Tab for which annotations are being loaded
 */
AnnozillaIconManager.prototype.userLoadBeginning = function(target) {
    this.removeIcons(target);
}

/**
 * Called when a 'check for replies' operation is beginning.
 * @param uri URI of the annotation for which to check replies.
 * @param target Tab with which the annotation is associated.
 */
AnnozillaIconManager.prototype.replyCheckBeginning = function(uri, target) {
    // Nothing to do here
}
/**
 * Inform the Icon Manager of the progress of an operation.
 * @param serversRemaining Number of servers yet to respond
 * (i.e. 0 means operation complete).
 * @param target Target of the operation
 */
AnnozillaIconManager.prototype.updateProgress =
function(serversRemaining, target) {
}

// called when either insertIcons or insertHighlight prefs are set 
AnnozillaIconManager.prototype._insertPrefChange = function() {
    var insert = this.insertIcons || this.insertHighlight;
    if (insert && (! this.observing) ) {
	// add our observer
        this.annozilla.app.addObserver(this);
	this.observing = true;
    }
    if ((!insert) && this.observing) {
	// remove our observer
	this.annozilla.app.removeObserver(this);
	this.observing = false;
    }
}  

/**
 * Called when a new annotation is going to be posted.
 */
AnnozillaIconManager.prototype.postBeginning = function (target)
{
}

/**
 * Success callback function to return the results for a loadAnnotations request.
 * This function returns the annotations for one server, designated by serverURI.  
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI that
 *                    were on serverURI annotations server
 */
AnnozillaIconManager.prototype.loadAnnotationsOnSuccess =
function(target, serverURI, annotatedURI, annotationsLength, annotations) {
    //NOOP: wait for Annozilla to call us to load in order
}

/**
 * Callback function to indicate failure for a loadAnnotations request.
 * This function indicates failure for one particular annotations server,#
 * designated by serverURI.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */
AnnozillaIconManager.prototype.loadAnnotationsOnFailure =
function(serverURI, annotatedURI, message) {}

/**
 * Called when an annotation post request succeeds.
 *
 * Only needs to do xpointer resolution if _isInsertIconsEnabled
 *   (Annozilla..postAnnotationOnSuccess performs xpointer resolution 
 *    independantly and redundantly in any case before adding to sidebar)
 *
 * @param newAnnotation nsIAnnotation the newly created annotation
 */
AnnozillaIconManager.prototype.postAnnotationOnSuccess =
function(newAnnotation, target, postWindow) {
    //NOOP: let Annozilla call addAnnotation
    // though it wouldn't be too bad to resolve redundantly below in this case
}

/**
 * Called when an annotation post fails.
 *
 * @param message string error message
 */
AnnozillaIconManager.prototype.postAnnotationOnFailure =
function(message, target, postWindow) {}

/**
 * Called when an existing annotation is going to be updated.
 */
AnnozillaIconManager.prototype.updateBeginning = function(uri, target) {
}

/**
 * Called when an annotation update succeeds.
 *
 * @param updatedAnnotation nsIAnnotation that has been updated on server
 */
AnnozillaIconManager.prototype.updateAnnotationOnSuccess =
function(updatedAnnotation, target, updateWindow) {

    if (this._isInsertIconsEnabled()) {

        var range = this.annozilla.resolveAnnotation(updatedAnnotation, target);
        this._insertAnnotationBefore(updatedAnnotation, range, target.contentDocument);
        /*
        // update the type, if there is an icon, and it is displaying
        var img = this.icons[updatedAnnotation.annotationURI];
        if (img && img.parentNode) {
            var type = updatedAnnotation.data.getSecondaryType();
            img.setAttribute("src", this._getIconURI(type));
            img.setAttribute("alt", type);
        }
        */

    }
}

/**
 * Called when an annotation update fails.
 *
 * @param url string url of the annotation that could not be updated
 * @param message string error message
 */
AnnozillaIconManager.prototype.updateAnnotationOnFailure =
function(url, message, target, updateWindow) {}

/**
 * Called when an annotation is successfully deleted.
 *
 * @param url string url of deleted annotation
 */
AnnozillaIconManager.prototype.deleteAnnotationOnSuccess = function(uri, target) {

    if (this._isInsertIconsEnabled()) {

        annozillaHighlighter.removeHighlight(uri, target);
        
        var model = new AnnozillaIconDataModel (target);
        var icons = model.getIcons();
        
        var img = icons[uri];
        if (img) {
            // if there is an img, remove it
            this._removeIcon(img);
        } else {
            dump("No icon to delete?\n");
        }
    }
}

/**
 * Called when an annotation delete fails.
 *
 * @param url string url of the annotation that could not be deleted
 * @param message string error message
 */
AnnozillaIconManager.prototype.deleteAnnotationOnFailure = function(url, message, target) {}

/**
 * Called when the replies to an annotation have been successfully
 * loaded.
 * 
 * @param annotation nsIAnnotation for which replies were requested
 * @param repliesEnum nsISimpleEnumerator containing nsIAnnotation replies
 */
AnnozillaIconManager.prototype.getRepliesOnSuccess = function(annotation, repliesCount, replies, target) {}

/**
 * Called when getting the replies to an annotation fails.
 *
 * @param annotation nsIAnnotation for which replies were requested
 * @param message string error message
 */
AnnozillaIconManager.prototype.getRepliesOnFailure = function(annotation, message, target) {}

/**
 * Called an annotation has successfully been replied to.
 * 
 * @param original nsIAnnotation that existed originally
 * @param reply nsIAnnotation reply created
 */
AnnozillaIconManager.prototype.replyToAnnotationOnSuccess =
function(original, reply, target, replyWindow) {}

/**
 * Called when replying to an annotation fails.
 *
 * @param original nsIAnnotation to which we attempted to reply
 * @param message string error message
 */
AnnozillaIconManager.prototype.replyToAnnotationOnFailure =
function(original, message, target, replyWindow) {}

/**
 * Called after a loadAnnotationBody request
 * when the body text of the annotation has been successfully loaded.
 *
 * @param annotation nsIAnnotation with updated body text
 * @param bodyText string text of the annotation's body
 */
AnnozillaIconManager.prototype.loadAnnotationBodyOnSuccess = function(annotation, bodyText) {}

/**
 * Called after a loadAnnotationBody request when
 * the body text fails to load.
 *
 * @param annotation nsIAnnotation that body was requested for (unmodified)
 * @param message string error message
 */
AnnozillaIconManager.prototype.loadAnnotationBodyOnFailure = function(annotation, message) {}

AnnozillaIconManager.prototype.iconClick = function(annotation) {
    if(annotation) {
        this.annozilla.update(annotation.annotationURI, gBrowser.selectedBrowser);
    }
}

AnnozillaIconManager.prototype.addIconForEditorWindow = function (win, fragment, target) {
    var range = this.annozilla.resolveFragment(fragment, target);
    if(!range) {
        throw "resolveFragment failed for fragment '" + fragment + "'";
    }
    var icon  = this._insertIcon(null, range, target.contentDocument, target);
    var im = this;
    win.addEventListener(
        "unload",
        function(e){
            if(win.closed) im._removeIcon(icon);
        },
        false
    );
};

AnnozillaIconManager.prototype.removeIcons = function(target) {
    var model = new AnnozillaIconDataModel(target);
    var icons = model.getIcons();
    for (uri in icons) {

        annozillaHighlighter.removeHighlight(uri, target);
        
        var img = icons[uri];
        if (img)
        {
            this._removeIcon(img);
        }
    }
    model.clearIcons();
}

/**
 * Called during load (by Annozilla..addAnnotations/_AddNext)  
 *   with xpointer resolution already taken care of
 */
AnnozillaIconManager.prototype.addAnnotation = function(annotation, range, doc, target) {
    if (this._isInsertIconsEnabled() && range) {    
        var img = this._insertAnnotationBefore(annotation, range, doc);
        this._addIconToIcons(img, annotation.annotationURI, target);
    }
}

/**
 * Called when a reply is created.
 * No icons are displayed for replies.
 */
AnnozillaIconManager.prototype.addReply =
function (inReplyTo, reply, range, doc, target)
{
}

/********************
 * PRIVATE FUNCTIONS
 ********************/

AnnozillaIconManager._hasRequiredComponents = function() {
    return true;
}

// Add the icon icon to the collection of icons for this tab.
AnnozillaIconManager.prototype._addIconToIcons = function(icon, annotationURI, target){
    var model = new AnnozillaIconDataModel(target);
    var icons = model.getIcons();
    try {
        icons[annotationURI] = icon;
        model.setIcons(icons);
    } catch (e) {
        dump("Where'd the icons go?\n  "+e+"\n");
    }
};

AnnozillaIconManager.prototype._iconForAnnotation = function(annotation, type, doc) {
    // create the icon
    var img = doc.createElementNS(
        Annozilla.XHTML_NS,
        "img"
    );
    if(annotation) {
        img.setAttribute("src", this._getIconURI(type));
        img.setAttribute("class", "annozillaAnnotationIcon"); 
    } else {
        img.setAttribute("src", "chrome://annozilla/content/icons/newAnnotation.png");
        img.setAttribute("class", "annozillaUnsavedAnnotationIcon"); 
    }

    return img;
}

AnnozillaIconManager.prototype._insertIconBefore = function(annotation, range, icon) {
    // icon is now added to icons in _insertIcon.

    // mark the element to be ignored by XPointer resolution
    gAnnozillaXPointerUtils.xptrMarkElement(icon);

    AnnozillaIconManager._insertNode(range, icon);

    var that = this;

    icon.addEventListener(
        "click",
        function(e){
            //prevent link action when the icon is inside the anchor <a> area.
            e.preventDefault();
            that.iconClick(annotation);
        },
        false
    );

    return icon;
}

/**
 *  Inserts highlighting span and/or icon image
 *  Unless such elements are already found with associated annotation-id attribute
 *  - this allows for cached page or reload of sbar but requires page
 *  to be reloaded or elements removed to refresh 
 *  (annotation-id attribute may be useful for removal/edit)
 *
 *  - ideally we'd want some sort of checksum to make sure annotation length
 *    hasn't changed?
 *
 *  Also note:
 *   annotation.data "type" property of "type#subtype" 
 *     allows new type to share a particular icon subtype style? (experimental)
 *
 *   new icon given spanlength attribute to allow for proper nesting
 *
 */
AnnozillaIconManager.prototype._insertAnnotationBefore = function(annotation, range, doc) {
    if (!range) throw "a range is required for _insertAnnotationBefore";

    var uri = annotation.annotationURI;
    var type = annotation.data.getSecondaryType();

    //first see whether there are already icons or spans for this 
    var img;
    //the parent of the start container may be an existing highlight span/div
    var span = range.startContainer.parentNode;

    //highlights insert page annotations just under html
    // so instead of searching up the tree, need to search forward
    // starting with first child of the html node
    //also check page annotation body for a stylesheet directive
    var seekForward = false;
    if (range.startContainer.nodeName.toLowerCase() == "html") {
        //dump("[Page Annot]");
        span = range.startContainer.firstChild;
        seekForward = true;
    }

    //and if we're just inside a document, we'd better move inside
    // this can happen when an annotation was created without a selection
    // annotations really ought to be restricted to body elements?
    if (span.nodeType == span.DOCUMENT_NODE) {
        span = range.startContainer.firstChild;
    }

    // highlighted spans might be nested
    while (span) {
        if (span.nodeType == span.ELEMENT_NODE) {
            //&& (span.hasAttribute("_dom_ignore_element_"))) {
            // first, get above wrapped elements
            if ((span.hasAttribute("_dom_ignore_element_"))
                && (span.getAttribute(AnnozillaIconManager.ANNOTATION_ID_ATTRIBUTE_NAME) == uri)
                && ((span.nodeName == "span")
                    || (span.nodeName == "div"))) {
                //typically, we assume it is our span or div
                // but if seekForward it may be an img so we need to check
                break;
            } else {
                if (seekForward) {
                    //page annotations are inserted one after another
                    span = span.nextSibling;
                } else {
                    // keep going up
                    //  assuming for now that icons don't appear without spans,
                    //  else we might have to look at previousSibling
                    span = span.parentNode; 
                }
            }
        } else {
            span = null;
        }
    }        

    // highlight first because _insertNode can alter dom path to end of range
    if (this.insertHighlight) {
        if (span) {
            //update existing span
            span.setAttribute(AnnozillaIconManager.ANNOTATION_TYPE_ATTRIBUTE_NAME, type);
        } else {
            //insert span/div 
            //  with class, annotation-type, annotation-id, segment attributes
            annozillaHighlighter.addHighlight(range, type, annotation);
            //move back one to insert icon before span
            if (range.startOffset > 0) {
                range.setStart(range.startContainer,range.startOffset-1);
            }
        } 
    }

    if (this.insertIcons) {
        // now look where there should be an icon
        if (span) {
            img = span.previousSibling;
        } else {
            img = range.startContainer.previousSibling;
        }
        
        while (img) {
            if ((img.nodeType == img.ELEMENT_NODE)
                && (img.nodeName == "img")
                && (img.getAttribute(AnnozillaIconManager.ANNOTATION_ID_ATTRIBUTE_NAME) == uri)) {
               //update existing icon
               img.setAttribute(AnnozillaIconManager.ANNOTATION_TYPE_ATTRIBUTE_NAME, type);

               img.setAttribute("src", this._getIconURI(type));
               img.setAttribute("alt", type);               
               return img; 
             } else {
                 //might be a footnote or another img?
                 img = img.previousSibling;
             }
        }
        // else... new icon should go before existing span
        if (span) {
            range.setStartBefore(span);
        }
        
        img = this._iconForAnnotation(annotation, type, doc);

        // mark the element to be ignored by XPointer resolution
        gAnnozillaXPointerUtils.xptrMarkElement(img);

        img.setAttribute(AnnozillaIconManager.ANNOTATION_ID_ATTRIBUTE_NAME, uri);
        //allow selection for CSS, XBL (experimental)
        img.setAttribute(AnnozillaIconManager.ANNOTATION_TYPE_ATTRIBUTE_NAME, type); 
        // spanlength allows highlight to determine proper nesting
        img.setAttribute("spanlength", range.toString().length);

        img = this._insertIconBefore(annotation, range, img);
        
    };
    return img;

}

// This function adds an icon for annotation annotation at the start of range
// range within the specified document, or adds a
// temporary icon (for an as yet unsaved annotation) if annotation is null.
AnnozillaIconManager.prototype._insertIcon = function(annotation, range, doc, target) {
    var icon;
    if(range == null) {
        // Insert the icon at the start of the document.
        // (this is not used (or tested) because the two places that call
        // _insertIcon both check that range is not null).
        icon = this._insertAnnotationBefore(
            annotation,
            doc.body.firstChild,
            doc
        );
    } else {
        icon = this._insertAnnotationBefore(annotation, range, doc);
    }
    if(annotation) { this._addIconToIcons(icon, annotation.annotationURI, target); }
    return icon;
}

/**
 * Get the URL for the icon associated with an annotation type.
 * @param type Full URI of the type identifying the annotation type
 * (the so-called "secondary type").
 * @return URI of the icon associated with the type
 */
AnnozillaIconManager.prototype._getIconURI = function(type) {  

  var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);
    
  var r = this.annozilla.annotationTypes.GetSource
           (rdfService.GetResource (Annozilla.NC+'URL'),
            rdfService.GetLiteral(type),
            true);
  if (r) {
      var i = this.annozilla.annotationTypes.GetTarget
          (r,
           rdfService.GetResource(Annozilla.NC+'Icon'),
           true);
      if (i && i.QueryInterface(Components.interfaces.nsIRDFLiteral)) {
           return i.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
      }
  }
  return 'chrome://annozilla/content/icons/annot.png';
}

/**
 * Inserts a node into the Document or DocumentFragment at the start
 * of the Range. If the container is a Text node, this will be split
 * at the start of the Range (as if the Text node's splitText method
 * was performed at the insertion point) and the insertion will occur
 * between the two resulting Text nodes. Adjacent Text nodes will not
 * be automatically merged. If the node to be inserted is a
 * DocumentFragment node, the children will be inserted rather than
 * the DocumentFragment node itself.
 *
 * @param range DOM Range Object
 * @param node DOM Node to insert
 */
AnnozillaIconManager._insertNode = function(range, node_to_insert) {
  var ref_node;   // node beyond insertion point
  var parent;     // parent node

  if (range.startContainer.nodeType == node_to_insert.TEXT_NODE || 
      range.startContainer.nodeType == node_to_insert.CDATA_SECTION_NODE) {
    ref_node = range.startContainer.splitText(range.startOffset);
    
    range.startContainer.parentNode.insertBefore(node_to_insert, ref_node);
    return;
  }    
  
  // if the insertion point is at the end of the startContainer node just 
  // append the node to the startContainer node.
  if ( range.startOffset == range.startContainer.childNodes.length) {
    range.startContainer.appendChild( node_to_insert );
  }
  // otherwise, insertion point is not at end of startContainer
  else {
    range.startContainer.insertBefore( node_to_insert,
				       range.startContainer.childNodes.item(range.startOffset));
  }
}

AnnozillaIconManager.prototype._removeIcon = function(icon) {
    if (icon.parentNode) {
        icon.parentNode.removeChild(icon);
    }
}

// _isInsertIcons[OrHighlight]Enabled
AnnozillaIconManager.prototype._isInsertIconsEnabled = function() {
    return this.insertIcons || this.insertHighlight;
}

AnnozillaIconManager.prototype.noRepliesFor = function (annotation, target)
{
}

AnnozillaIconManager.prototype.tabSelected = function (browser) {
    var icons = new AnnozillaIconDataModel (browser);
    if (icons.getIcons() == null) {
        icons.clearIcons();
    }
}
