/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Annozilla.
 *
 * The Initial Developer of the Original Code is
 * Matthew Wilson <matthew@mjwilson.demon.co.uk>.
 * Portions created by the Initial Developer are Copyright (C) 2002
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Eliot Setzer
 *   C. Greg Hagerty <cgreg@cgreg.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const ANNOZILLA_XUL_NS = "http://annozilla.mozdev.org/xul/2005";

// Each window has a single gAnnozilla instance.
// Annotation data itself is held in a data model associated with
// each tab; see model.js.
var gAnnozilla; // was set by initialize.js, now by initAnnozilla via overlay
var gAnnozillaBrowser; // for tab handling

function initAnnozilla () {
    gAnnozilla = new Annozilla();
    window.addEventListener("unload", 
                            function (e) {gAnnozilla.unload();},
                            false);

    gAnnozillaBrowser = window.top.gBrowser;
    var container = gAnnozillaBrowser.tabContainer;
    container.addEventListener("TabSelect", gAnnozilla.tabSelected, false);
    gBrowser.addEventListener("DOMContentLoaded", gAnnozilla.contentLoaded, true);

}

function Annozilla() {
    
    if (! Annozilla._hasRequiredComponents()) {
        return;
    }
    
    this.windowIndex = 0;

    this.bundle = srGetStrBundle("chrome://annozilla/locale/annozilla.properties");

    // initialize annotations service
    this.annotService =
       Components.classes[Annozilla.ANNOTATIONS_SERVICE_CONTRACTID].getService().
           QueryInterface(Components.interfaces.nsIAnnotationsService);

    // get the annozilla AnnotationsApplication object
    this.app = this.annotService.registerApplication("annozilla");

    // add our observer
    this.app.addObserver(this);

    // icons module
    this.icons = new AnnozillaIconManager(this);
    // sidebar module, when loaded
    this.sidebar = null;
    this.views = new AnnozillaViews;
    this.views.addView (this.icons);

    // prefs module
    this.prefs = new AnnozillaPrefs(this);

    // This is necessary to prevent _addAnnotationToContainer from failing when
    // an annotation is created before any annotations have ever been loaded by
    // the sidebar. TODO: only deals with the current tab
    // Action: delete this and ensure that the model handles this scenario.
    //this.clearAnnotations(gBrowser.selectedBrowser);

    this.defaultBodyDestination = window.document.webNavigation;

    // RDF datasource which can be used to add new annotation types
    this.annotationTypes =
       Components.classes["@mozilla.org/rdf/datasource;1?name=composite-datasource"].
            createInstance(Components.interfaces.nsIRDFCompositeDataSource);
    this.annotationTypes.AddDataSource
       (Components.classes["@mozilla.org/rdf/rdf-service;1"].
            getService(Components.interfaces.nsIRDFService).
            GetDataSourceBlocking("chrome://annozilla/content/resources/annotationTypes.rdf"));
    /*
    this.annotationTypes.AddDataSource
       (Components.classes["@mozilla.org/rdf/rdf-service;1"].
            getService(Components.interfaces.nsIRDFService).
            GetDataSourceBlocking("chrome://annozilla/content/resources/sample.rdf"));
    */
}

Annozilla.UNKNOWN_REASON_URL = "about:blank";

Annozilla.ANNOTATIONS_SERVICE_CONTRACTID = "@mozilla.org/annotations_service;1";
Annozilla.XPOINTER_SERVICE_CONTRACTID = "@mozilla.org/xpointer-service;1";

Annozilla.ANNOT_NS = "http://www.w3.org/2000/10/annotation-ns#";
Annozilla.RDF_NS   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
Annozilla.DC10_NS    = "http://purl.org/dc/elements/1.0/";
Annozilla.DC11_NS    = "http://purl.org/dc/elements/1.1/";
Annozilla.THREAD_NS= "http://www.w3.org/2001/03/thread#";
Annozilla.XHTML_NS = "http://www.w3.org/1999/xhtml";
Annozilla.TYPE_NS = "http://www.w3.org/2000/10/annotationType#";

Annozilla.NC = "http://home.netscape.com/NC-rdf#";
Annozilla.ATYPES = 'chrome://annozilla/content/resources/annotationTypes.rdf';
Annozilla.DEFAULT_ICON_IMG = 'chrome://annozilla/content/icons/post-it.png';

// The previous value for this ("annotation-body") produced warnings (about the
// '-' not being a legal character) in a version of Firefox compiled with
// debugging messages.
Annozilla.WINDOW_ID_PREFIX = "annotationBody";

Annozilla.prototype = {
    windowIndex: undefined,
    bundle: undefined,
    defaultBodyDestination: undefined
};

Annozilla.prototype.logToConsole = function (message)
{
    Components.classes['@mozilla.org/consoleservice;1'].
        getService(Components.interfaces.nsIConsoleService).
          logStringMessage (message);
};

Annozilla.prototype.alert = function (title, message)
{
    const promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].
        getService(Components.interfaces.nsIPromptService);
    promptService.alert (window, title, message);
};

/**
 * Add a view to the list of the views onto the annotations.
 */
Annozilla.prototype.addView = function (view)
{
    this.views.addView (view);
}

/**
 * Remove a view from the list of the views onto the annotations.
 */
Annozilla.prototype.removeView = function (view)
{
    this.views.removeView (view);
}

Annozilla.prototype.updateProgressViews = function (serverCount, target) {
    this.views.updateProgressViews (serverCount, target);
}

Annozilla.prototype.clearAnnotations = function (target)
{
    new AnnozillaAnnotationModel (target).clearAnnotations();
};

Annozilla.prototype.unload = function() {
    // remove observer
    this.app.removeObserver(this);

    // tell prefs we're shutting down
    this.prefs.destructor();
    
    var container = gAnnozillaBrowser.tabContainer;
    container.removeEventListener("TabSelect", gAnnozilla.tabSelected, false);
    
    gBrowser.removeEventListener("DOMContentLoaded", gAnnozilla.contentLoaded, true);

};

Annozilla.prototype.QueryInterface = function(iid) {
    if (! (iid.equals(Components.interfaces.nsIAnnotationsObserver) ||
           iid.equals(Components.interfaces.nsISupports) ) ) {
        throw Components.results.NS_ERROR_NO_INTERFACE;
    }
    return this;
};

Annozilla.prototype.insertIconsPrefChanging = function(icons){
    this.views.insertIconsPrefChanging(icons);
};

Annozilla.prototype.insertHighlightPrefChanging = function(highlight){
    this.views.insertHighlightPrefChanging(highlight);
};

// Return the range specified by the fragment.
Annozilla.prototype.resolveFragment =
function(fragment, target) {
    return gAnnozillaXPointerUtils.xptrResolveFragment(fragment, target.contentDocument);
};

// Return the range to which the annotation applies.
Annozilla.prototype.resolveAnnotation =
function(annotation, target) {
    return gAnnozillaXPointerUtils.xptrResolveAnnotation(annotation, target.contentDocument);
};

/**
 * Notification that a document has finished loading and constructed its DOM.
 * If autoloading is enabled, then start loading annotations for the page.
 */
Annozilla.prototype.contentLoaded = function (event)
{
    // TODO This seems to fail when pages are restored at startup.    
    var url = event.currentTarget.mCurrentBrowser.contentDocument.location.href;
    
    // TODO frame detection? non-HTML doc types?
    if (gAnnozilla.autoLoadAllowSite(url)) {
        gAnnozilla.userLoadOnCommand (event.currentTarget.mCurrentBrowser);
    }
};

// Determine whether an annotation is associated with a Local Server
// (Should be a method on the annotation?)
Annozilla.prototype._isLocal = function (annotation)
{
    return annotation.serverName == this.app.LOCALSERVER_URI;
};

// Generate a URI for the body of an annotation.
// For a local annotation, create a data: URI for the body.
// For a remote annotation, just use the existing URI for the body.
Annozilla.prototype._generateURIForAnnotation = function (annotation)
{
    if (this._isLocal (annotation)) {
        return LocalServerUtils.createTempURIFromString(annotation.getBodyText());
    } else {
        return annotation.bodyURI;
    }
};

/*
 * Currently, annotation collection is cleared and reloaded for each tab switch
 * But _addAnnotations/_addNextAnnotation checks each annotation URI is for current page
 *  so that if we want to avoid reload we could accumulate annotations
 *  for open pages by not resetting the collection
 */
Annozilla.prototype.tabSelected = function (event) {

    var browser = gAnnozillaBrowser.selectedBrowser;
    // browser is the newly selected tab

    // Why not "this.views"? Because "this" refers to an
    // XUL element when called.
    gAnnozilla.views.tabSelected (browser);
};

/**
 * Return the chrome URL used for editing annotation bodies.
 */
Annozilla.prototype.getEditorUrl = function () {
    return "chrome://annozilla/content/editor/editor.xul";
};

/**
 * Loads annotations for the current page.
 *
 * @param target Optional argument indicating the browser
 * in which the load occurs.
 */   
Annozilla.prototype.userLoadOnCommand = function(target) {
    
    // If no target was specified, get the selected tab.    
    if (target == undefined)
    {
        target = gBrowser.selectedBrowser;
    }
    
    //this.userLoading = true;
    this.views.userLoadBeginning(target);

    // Finish if nothing to do
    if (this.currentServerCount() === 0) {
        return;
    }

    this.clearAnnotations(target);
    
    var model = new AnnozillaAnnotationModel (target);
    var serverCount = model.startOperation (this.currentServerCount(), true);

    this.updateProgressViews (serverCount, target);

    // Do the load. We will be notified of the results by callbacks.
    this.app.loadAnnotations (
        this.stripURL(target.contentDocument.location.href),
        gBrowser.selectedBrowser);

};

Annozilla.prototype.checkForRepliesOnCommand = function(uri, target) {

    var model = new AnnozillaAnnotationModel (target);

    if (!model.hasAnnotation(uri))
    {
        this.alert ("Annozilla", this.bundle.GetStringFromName ("INTERNAL_ERROR"));
        return;
    }
    
    this.views.replyCheckBeginning(uri, target);

    var serverCount = model.startOperation (1); // We only check the server on which the annotation is saved
    this.updateProgressViews (serverCount, target);

    var annotation = model.getAnnotation(uri).annot;

    this.app.getReplies (annotation, gBrowser.selectedBrowser);
};

Annozilla.prototype.openEditorWindow = function(
    windowOptions,
    annotation,
    baseURL,
    fragment,
    bodyURI,
    inReplyTo,
    inReplyToBodyURI,
    threadRoot,
    target) {
    
    var w = window.openDialog(
        this.getEditorUrl(),
        Annozilla.WINDOW_ID_PREFIX + this.windowIndex++,
        windowOptions,
        annotation,
        baseURL,
        fragment,
        bodyURI,
        inReplyTo,
        inReplyToBodyURI,
        threadRoot,
        target);
    
    if(fragment) {
        // Fails because the icon manager requires an annotation
        // and we don't have one
        this.icons.addIconForEditorWindow(w, fragment, target);
    } else if (annotation == undefined) {
        throw("openEditorWindow requires either a fragment or an existing annotation");
    }
};

Annozilla.prototype.post = function(baseURL, fragment, target){
    this.openEditorWindow(
        "chrome,resizable,dialog=no",
        null, // annotation
        baseURL,
        fragment,
        null, // bodyURI,
        null, // inReplyTo
        null, // inReplyToBodyURI
        null, // threadRoot
        target);
};

Annozilla.prototype.postOnCommand = function() {
    // get the xptr
    // TODO assumes the current tab, is this OK?
    var seln = window._content.getSelection();

    /*
    if (seln.anchorNode == null) {
        // it might be a good idea to move into html[1]/body[1]? if it exists.
    }
    */
    
    var baseURL = this.stripURL(window._content.location.href);
    var xptr = gAnnozillaXPointerUtils.xptrCreateXPointer(seln, window._content.document);

    var target = window.top.gBrowser.selectedBrowser;
    this.views.postBeginning (target);

    this.post(baseURL, xptr, gBrowser.selectedBrowser);
};

Annozilla.prototype.replyOnCommand = function(replyToURI) {
};

Annozilla.prototype.update = function(annotationURI, target) {
    
    var model = new AnnozillaAnnotationModel (target);
    
    if (! model.hasAnnotation(annotationURI) ) {
        this.alert("Annozilla", this.bundle.GetStringFromName ("EDIT_INTERNAL_ERROR"));
        return;
    }
    
    this.views.updateBeginning(annotationURI, target);  

    var annotation = model.getAnnotation(annotationURI);
    var bodyURI = this._generateURIForAnnotation(annotation.annot);

    // If the annotation has a thread root set, use it.
    // Otherwise, we are editing an annotation, not a reply,
    // so use this annotation's URI as the root if we create any replies.
    this.openEditorWindow(
        "chrome,resizable,dialog=no",
        annotation.annot.clone(),
        null,
        null,
        bodyURI,
        annotation.inReplyTo,
        null, // inReplyToBodyURI
        annotation.threadRoot,
        target);
};

Annozilla.prototype.updateOnCommand = function(annotationURI, target) {
    this.update(annotationURI, target);
};

Annozilla.prototype.deleteOnCommand = function(annotationURI, target) {

    var model = new AnnozillaAnnotationModel (target);
   
    if (! model.hasAnnotation(annotationURI)) {
        this.alert("Annozilla", this.bundle.GetStringFromName ("DELETE_INTERNAL_ERROR"));
        return;
    }

    var serverCount = model.startOperation (1); // always just 1 server for delete
    this.updateProgressViews (serverCount, target);

    var annotation = model.getAnnotation(annotationURI).annot;
    this.app.deleteAnnotation(annotation, target);

};

Annozilla.prototype.loadBodyOnCommand = function(annotationURI, target) {
    
    var model = new AnnozillaAnnotationModel (target);
    
    if (! model.hasAnnotation(annotationURI)) {
        this.alert("Annozilla", this.bundle.GetStringFromName ("LOAD_INTERNAL_ERROR"));
        return;
    }
  
    var annotation = model.getAnnotation(annotationURI).annot;
    
    var bodyURI = this._generateURIForAnnotation(annotation);
  
    if (this.prefs.getPref("loadInNewWindow")) {
        window.open(bodyURI, "_blank","width=600,height=320,scrollbars=yes,resizable=yes","");
    }
    else {
        this.loadAnnotationBody(bodyURI);
    }
};

Annozilla.prototype.autoLoadAllowSite = function(url){
    if (this.prefs.getPref("autoLoadMulti") == "never") {
        return false;
    }
    if (this.prefs.getPref("autoLoadMulti") == "sites") {
        var allowSite = false;
        var sites = this.prefs.getPref("autoLoadSites").split(" ");
        for (var ix=0; ix<sites.length; ix++) {
            if (url.indexOf(sites[ix]) === 0) {
                allowSite = true;
                dump("Autoload allowed for "+url+"\n");
                break;
            }
        }
        if (!allowSite) {
            return false;
        }
    }
    return true;
};

Annozilla.prototype.byeByeSidebar = function () {
    dump("Annozilla: Bye bye Sidebar\n");
    this.sidebar = null;
};

/**
 * Success callback function to return the results for a loadAnnotations request.
 * This function returns the annotations for one server, designated by serverURI.  
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri that the annotations annotate
 * @param annotations nsISimpleEnumerator of nsIAnnotations for annotatedURI
 *                    that were on serverURI annotations server
 *  @param target Target of the load
 */
Annozilla.prototype.loadAnnotationsOnSuccess =
function (serverURI, annotatedURI, annotationsLength, annotations, target) {

    try {

        var model = new AnnozillaAnnotationModel (target);
        var serverCount = model.serverResponded();
        
        // If we've moved on to a new page, ignore these annotations.
        // Work in progress: it's OK if we've moved on to another tab,
        // just load the annotations in the correct tab
        if (annotatedURI == target.currentURI.spec) {
            this._addAnnotations (annotations, target);
        } else {
            dump(
                 "ignoring loadAnnotationsOnSuccess of " + annotatedURI + " from " + 
                 serverURI + " because it pertains to old URI.  currentURI= " +
                 target.currentURI.spec + "\n");
        }

        this.updateProgressViews (serverCount, target);


    } catch (e) {
        var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
        var console = consoleService.getService(Components.interfaces.nsIConsoleService);
        console.logStringMessage("Annozilla: error loading annotations");
        Components.utils.reportError (e);
    }

};

/**
 * Callback function to indicate failure for a loadAnnotations request.
 * This function indicates failure for one particular annotations server,
 * designated by serverURI.
 *
 * @param serverURI string uri of the annotations server responding
 * @param annotatedURI string uri for which annotations were requested
 * @param message string error message
 */
Annozilla.prototype.loadAnnotationsOnFailure = function(serverURI, annotatedURI, message, target) {
    Components.classes['@mozilla.org/consoleservice;1'].
      getService(Components.interfaces.nsIConsoleService).
      logStringMessage
        ("Annozilla: failed to load annotations from: " + serverURI +
         " because of error: " + message + "\n");
    
    try {
        var model = new AnnozillaAnnotationModel (target);
        var serverCount = model.serverResponded();
        this.updateProgressViews (serverCount, target);
    } catch (e) {
        var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
        var console = consoleService.getService(Components.interfaces.nsIConsoleService);
        console.logStringMessage("Annozilla: error loading annotations");
        Components.utils.reportError (e);
    }
};

/**
 * Called when an annotation post request succeeds.
 *
 * @param newAnnotation nsIAnnotation the newly created annotation
 */
Annozilla.prototype.postAnnotationOnSuccess = function(newAnnotation, target, postWindow) {

    // if this annotation applies to a page we're no longer viewing
    var uri = newAnnotation.data.getProperty(Annozilla.ANNOT_NS, "annotates");
    if (this.stripURL(uri) != this.stripURL(target.contentDocument.location.href)) {
        dump("posted annotation applies to page no longer being viewed\n");
        dump(" ('"+uri+"' while viewing '"+target.contentDocument.location.href+"')\n");
        return;
    }

    var range = this.resolveAnnotation(newAnnotation, target);

    new AnnozillaAnnotationModel (target).addAnnotation (newAnnotation, range);

    this.views.addAnnotation (newAnnotation, range, target.contentDocument, target);

};

/**
 * Called when an annotation post fails.
 *
 * @param message string error message
 */
Annozilla.prototype.postAnnotationOnFailure = function(message, target, postWindow) {};

/**
 * Called when an annotation update succeeds.
 *
 * @param updatedAnnotation nsIAnnotation that has been updated on server
 */
Annozilla.prototype.updateAnnotationOnSuccess =
function(updatedAnnotation, target, updateWindow) {
    
    var model = new AnnozillaAnnotationModel (target);
    model.updateAnnotation (updatedAnnotation);

};

/**
 * Called when an annotation update fails.
 *
 * @param url string url of the annotation that could not be updated
 * @param message string error message
 */
Annozilla.prototype.updateAnnotationOnFailure =
function(url, message, target, updateWindow) {};

/**
 * Called when an annotation is successfully deleted.
 *
 * @param url string url of deleted annotation
 */
Annozilla.prototype.deleteAnnotationOnSuccess = function(url, target) {
    
    var model = new AnnozillaAnnotationModel (target);
    var serverCount = model.serverResponded();
    
    model.deleteAnnotation(url);
    
    this.updateProgressViews (serverCount, target);

};

/**
 * Called when an annotation delete fails.
 *
 * @param url string url of the annotation that could not be deleted
 * @param message string error message
 */
Annozilla.prototype.deleteAnnotationOnFailure = function(url, message, target) {
    
    var model = new AnnozillaAnnotationModel (target);
    var serverCount = model.serverResponded();
    this.updateProgressViews (serverCount, target);

    this.alert ("Annozilla", this.bundle.formatStringFromName ("DELETE_FAIL_ERROR", [url, message], 2));
};

/**
 * Called when the replies to an annotation have been successfully
 * loaded.
 * 
 * @param annotation nsIAnnotation for which replies were requested
 * @param repliesCount Number of replies
 * @param replies Array of nsIAnnotation replies
 * @param target Target with which the replies are associated
 */
Annozilla.prototype.getRepliesOnSuccess =
function(annotation, repliesCount, replies, target) {
    // TODO If we've moved on to a new page, ignore these replies
    try {

        var model = new AnnozillaAnnotationModel (target);
        var serverCount = model.serverResponded();

        this._addReplies (annotation, repliesCount, replies, target);

        this.updateProgressViews (serverCount, target);


    } catch (e) {
        var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
        var console = consoleService.getService(Components.interfaces.nsIConsoleService);
        console.logStringMessage("Annozilla: error loading annotations");
        Components.utils.reportError (e);
    }
};

/**
 * Called when getting the replies to an annotation fails.
 *
 * @param annotation nsIAnnotation for which replies were requested
 * @param message string error message
 */
Annozilla.prototype.getRepliesOnFailure =
function(annotation, message, target) {
    var model = new AnnozillaAnnotationModel (target);
    var serverCount = model.serverResponded();
    this.updateProgressViews (serverCount, target);
};

/**
 * Called an annotation has successfully been replied to.
 * 
 * @param original nsIAnnotation that existed originally
 * @param reply nsIAnnotation reply created
 */
Annozilla.prototype.replyToAnnotationOnSuccess =
function(original, reply, target, replyWindow) {
    this._addReplies (original, 1, [ reply ], target);
};

/**
 * Called when replying to an annotation fails.
 *
 * @param original nsIAnnotation to which we attempted to reply
 * @param message string error message
 */
Annozilla.prototype.replyToAnnotationOnFailure =
function(original, message, target) {};

/**
 * Called after a loadAnnotationBody request
 * when the body text of the annotation has been successfully loaded.
 *
 * @param annotation nsIAnnotation with updated body text
 * @param bodyText string text of the annotation's body
 */
Annozilla.prototype.loadAnnotationBodyOnSuccess =
function(annotation, bodyText) {};

/**
 * Called after a loadAnnotationBody request when
 * the body text fails to load.
 *
 * @param annotation nsIAnnotation that body was requested for (unmodified)
 * @param message string error message
 */
Annozilla.prototype.loadAnnotationBodyOnFailure =
function(annotation, message) {};

Annozilla.prototype.setDefaultBodyDestination = function(destination){
    this.defaultBodyDestination = destination;
};

Annozilla.prototype.loadAnnotationBody = function(bodyURI, destination){
    if(destination) {
        destination.loadURI(bodyURI);
    } else {
        this.defaultBodyDestination.loadURI(bodyURI);
        this.defaultBodyDestination.docShell.allowJavascript = false;
        this.defaultBodyDestination.docShell.allowPlugins = false;
    }
};

Annozilla._hasRequiredComponents = function() {
    if (! Components.classes[Annozilla.ANNOTATIONS_SERVICE_CONTRACTID]) {
        return false;
    }
    if (! Components.classes[Annozilla.XPOINTER_SERVICE_CONTRACTID]) {
        return false;
    }
    return true;
};

/**
 * Private function to strip fragment ids off URLs.
 *
 * @param url String URL to strip fragment ids from
 * @return URL string stripped of fragment
 */
Annozilla.prototype.stripURL = function (url) {
    if(!url || ('string'!= typeof(url)) || ('' == url) || url == 'about:blank'){
        return 'about:blank';
    }

    return url.replace (/#.*/,'');
};

Annozilla.prototype.isAutoLoadEnabled = function() {
    return (this.prefs.getPref("autoLoadMulti") != "never");
};

Annozilla.prototype.currentServerCount = function(){
    return this.app.getServerCount();
};

Annozilla.prototype._addAnnotationToContainer = function (aAnnotation, aRange, aContainer) {
    aContainer[aAnnotation.annotationURI] = {
        annot: aAnnotation,
        range: aRange
    };
};

Annozilla.prototype._debugBadAnnotation = function(annotation) {
    dump("annotation.data is NULL.  enumerating properties:\n");
    dump("serverName: " + annotation.serverName + "\n");
    dump("annotationURI: " + annotation.annotationURI + "\n");
    dump("bodyURI: " + annotation.bodyURI + "\n");
    dump("getBodyText: " + annotation.getBodyText() + "\n");
    dump("setting bodyText.\n");
    annotation.setBodyText("New test body text");
    dump("new body text: " + annotation.getBodyText() + "\n");
};

Annozilla.prototype.userStopLoadCommand = function () {
    var servers = new AnnozillaAnnotationModel (gBrowser.selectedBrowser).stopResolving();
    if (servers == 0) {
        this.views.annotationResolutionFinished (gBrowser.selectedBrowser);
    }
    window._content.status = "Annotation loading stopped.";
};

Annozilla.prototype._addAnnotations = function (annotations, target) {
    
    var servers = new AnnozillaAnnotationModel (target).resolvingAnnotations();
    if (servers == 1) {
        this.views.annotationResolutionBeginning (target);
    }

    // If "target" is specified, then use it and get the document from
    // that tab. Otherwise use the main window's document.
    // TODO should we always have "target" set?
    var doc = (target == undefined) ? window._content.document : target.contentDocument;
    
    var self = this;
    setTimeout(function () { self._addNextAnnotation (0, annotations, doc, target); }, 0);

};

Annozilla.prototype._addNextAnnotation = function (i, annotList, doc, target) {
    var asynchLoading = new AnnozillaAnnotationModel (target).isResolving();
    if (asynchLoading && (i < annotList.length)) {
        try {
            var annot = annotList[i]; //consider popping this?
            
            try {
                if (! annot.data) {
                    this._debugBadAnnotation(current);
                }
                else
                {
                    var context = annot.data.getProperty(Annozilla.ANNOT_NS,"context");
                    var shortURI = doc.documentURI.lastIndexOf("/")+1;
                    shortURI = doc.documentURI.substr(shortURI); 
                    if (context.indexOf(shortURI) == -1) {
                        //SKIP ANNOTATIONS STORED FOR OTHER PAGES
                        dump("(skip)");
                    } else {
                        var range = gAnnozillaXPointerUtils.xptrResolveAnnotation(annot, doc);
                        var model = new AnnozillaAnnotationModel (target);
    
                        // add to the overall collection (annotations for all servers)
                        model.addAnnotation (annot, range);
                        this.addAnnotationType(annot.data.getSecondaryType());                        
                        this.views.addAnnotation(annot, range, doc, target);

                    }
                }
            } catch (e)
            {
                var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
                var console = consoleService.getService(Components.interfaces.nsIConsoleService);
                console.logStringMessage("Annozilla: error loading annotations");
                Components.utils.reportError (e);
            }

            var self = this;
            setTimeout(function () { self._addNextAnnotation (i+1, annotList, doc, target); }, 0);
            
        } catch (e) {
            var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
            var console = consoleService.getService(Components.interfaces.nsIConsoleService);
            console.logStringMessage("Annozilla: error loading annotations");
            Components.utils.reportError (e);
            //TODO: SIDEBAR
            //that.operationComplete("PAGE_LOAD_ERROR");
        }
    } else {
        //TODO: SIDEBAR
        //that.operationComplete("LOAD_OK");
        var servers = new AnnozillaAnnotationModel (target).resolvingComplete();
        if (servers == 0) {
            this.views.annotationResolutionFinished (target);
        }
        window._content.status = "Annotations loaded.";
        dump("\nDone!\n");
    }
};

// TODO can we unify _addAnnotations/_addReplies and
// _addNextAnnotation and _addNextReply?
Annozilla.prototype._addReplies = function (inReplyTo, repliesCount, replies, target) {
    
    var doc = target.contentDocument;

    if (repliesCount == 0)
    {
        this.views.noRepliesFor (inReplyTo, target);
    }
    else
    {
        var servers = new AnnozillaAnnotationModel (target).resolvingAnnotations();
        if (servers == 1) {
            this.views.annotationResolutionBeginning (target);
        }
        var self = this;
        setTimeout(function () { self._addNextReply (replies, doc, target); }, 0);
    }

};

// For _addNextReply it is tricky to find the correct
// order to items. We don't just have an array of replies
// to a single annotation, we have a complete reply tree,
// so as we iterate through, we may have a reply to an item
// that we haven't added yet.
Annozilla.prototype._addNextReply =
function (replyList, doc, target) {
    var asynchLoading = new AnnozillaAnnotationModel(target).isResolving();
    if (asynchLoading && (replyList.length > 0)) {
        try {
            var annot = replyList.shift();
            if (! annot.data) {
                this._debugBadAnnotation(current);
            }
            else
            {
                var context = annot.data.getProperty(Annozilla.ANNOT_NS,"context");
                /* TODO what should replace this?
                var shortURI = doc.documentURI.lastIndexOf("/")+1;
                shortURI = doc.documentURI.substr(shortURI); 
                if (context.indexOf(shortURI) == -1) {
                    //SKIP ANNOTATIONS STORED FOR OTHER PAGES
                    dump("(skip)");
                } else {
                */
                    var range = gAnnozillaXPointerUtils.xptrResolveAnnotation(annot, doc);
                    
                    var threadRoot =
                        annot.data.getResourceProperty(
                            Annozilla.THREAD_NS, "root");
                    
                    // This isn't necessarily the same as the
                    // inReplyTo argument, since 'getReplies'
                    // loads the whole tree!
                    var inReplyToReal =
                        annot.data.getResourceProperty(
                            Annozilla.THREAD_NS, "inReplyTo");
 
                    var model = new AnnozillaAnnotationModel (target);
                    
                    if ( ! model.hasAnnotation (inReplyToReal))
                    {
                        Components.classes['@mozilla.org/consoleservice;1'].
                            getService(Components.interfaces.nsIConsoleService).
                            logStringMessage ("Can't find " + inReplyToReal + " so abandoning " + annot.annotationURI);
                        replyList.push (annot);
                        // TODO how to avoid infinite loops?
                    }

                    // add to the overall collection (annotations for all servers)
                    model.addAnnotation (annot, range, inReplyToReal, threadRoot);
                    
                    this.views.addReply (inReplyToReal, annot, range, doc, target);

                /*
                }
                */
            }

            var self = this;
            setTimeout(function () { self._addNextReply (replyList, doc, target); }, 0);
            
        } catch (e) {
            var consoleService=Components.classes['@mozilla.org/consoleservice;1'];
            var console = consoleService.getService(Components.interfaces.nsIConsoleService);
            console.logStringMessage("Annozilla: error loading annotations");
            Components.utils.reportError (e);
            //TODO: SIDEBAR
            //that.operationComplete("PAGE_LOAD_ERROR");
        }
    } else {
        var servers = new AnnozillaAnnotationModel (target).resolvingComplete();
        if (servers == 0) {
            this.views.annotationResolutionFinished (target);
        }
        window._content.status = "Replies loaded.";
    }
};

// Open Firefox options. From extensions.js
Annozilla.prototype.openOptions = function () {

    var optionsURL = "chrome://annozilla/content/prefs/firefox/1.5/prefs-container.xul";
    var windows = Components.classes['@mozilla.org/appshell/window-mediator;1']
        .getService(Components.interfaces.nsIWindowMediator)
        .getEnumerator(null);
    while (windows.hasMoreElements()) {
        var win = windows.getNext();
        if (win.document.documentURI == optionsURL) {
            win.focus();
            return;
        }
    }

    var features;
    try {
        var instantApply =
            Components.classes["@mozilla.org/preferences-service;1"].
            getService(Components.interfaces.nsIPrefBranch).
            getBoolPref("browser.preferences.instantApply");
        features = "chrome,titlebar,toolbar,centerscreen" + (instantApply ? ",dialog=no" : ",modal");
    }
    catch (e) {
        features = "chrome,titlebar,toolbar,centerscreen,modal";
    }
    window.openDialog(optionsURL, "", features);

}

// From utilityOverlay.js goPreferences. Open Mozilla SeaMonkey preferences
Annozilla.prototype.openPreferences = function () {

    var containerID = "annozillaItem";
    var paneURL = "chrome://annozilla/content/prefs/pref-annozilla-viewing.xul";
    var itemID = "annozillaViewing";

    var resizable;
    var pref = Components.classes["@mozilla.org/preferences-service;1"]
        .getService(Components.interfaces.nsIPrefBranch);
    try {
        // We are resizable ONLY if in box debugging mode, because in
        // this special debug mode it is often impossible to see the 
        // content of the debug panel in order to disable debug mode.
        resizable = pref.getBoolPref("xul.debug.box");
    }
    catch (e) {
        resizable = false;
    }

    //check for an existing pref window and focus it; it's not application modal
    const kWindowMediatorContractID = "@mozilla.org/appshell/window-mediator;1";
    const kWindowMediatorIID = Components.interfaces.nsIWindowMediator;
    const kWindowMediator = Components.classes[kWindowMediatorContractID].getService(kWindowMediatorIID);
    var lastPrefWindow = kWindowMediator.getMostRecentWindow("mozilla:preferences");
    if (lastPrefWindow)
        lastPrefWindow.focus();
    else {
        var resizability = resizable ? "yes" : "no";
        var features = "chrome,titlebar,resizable=" + resizability;
        openDialog("chrome://communicator/content/pref/pref.xul","PrefWindow", 
                features, paneURL, containerID, itemID);
    }

}

/* 
 Add a temporary new annotationType to an in-memory copy 
  of the first annotation types datasource (for now)
  - maybe in the future the editor dropdown can merge datasources?
   given URI newType and optional args newName and newIcon
 otherwise, make a name that reflects it's dynamic
  - some extended feature may be able to show the URI's:
    eg. setting the editable attribute of the annotationType element
    in the editor at design time seems to show the URI's rather than names 
    but not so when toggled 
  @param newType URI identifying the annotation type
  @param newName (Optional) Textual description of the annotation type
  @param newIcon (Optional) URL of the icon to use for the type
*/
Annozilla.prototype.addAnnotationType =
function (newType, newName, newIcon) {

  var rdfService = Components.classes["@mozilla.org/rdf/rdf-service;1"]
    .getService(Components.interfaces.nsIRDFService);
    
  var gATypes = this.annotationTypes;
    
  if (newName == undefined) {
    // If the caller didn't pass a value for the name, construct one.
    // By default, we will try to use the fragment Id from the URL.
    newName = newType.substring(newType.indexOf("#")+1);
    // But if there's already an annotation type using this name,
    // use the full URL
    if (gATypes.GetSource
           (rdfService.GetResource (Annozilla.NC+'Name'),
            rdfService.GetLiteral(newName),
            true)) {
        newName = newType;
    }
  }
  newIcon = newIcon ? newIcon : Annozilla.DEFAULT_ICON_IMG;

  try {
    
    var rNewType = rdfService.GetResource(newType);
    
    // If there is a resource whose URL is that of the annotation type...
    if (gATypes.GetSource
           (rdfService.GetResource (Annozilla.NC+'URL'),
            rdfService.GetLiteral(newType),
            true)) {
        return;
    }
    
    var ds0 = gATypes.GetDataSources().getNext().QueryInterface(Components.interfaces.nsIRDFDataSource);

    if (ds0.URI != null) { // in-memory-datasource URI is null
        // otherwise, we need to convert to modify
        var ds = Components
            .classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
            .createInstance(Components.interfaces.nsIRDFDataSource);
        this.copyDataSource(ds0, ds);
        gATypes.RemoveDataSource(ds0);
        gATypes.AddDataSource(ds);
    } else {
        ds = ds0;
    }
    
    ds.Assert(rNewType,
              rdfService.GetResource(Annozilla.NC + 'Name'), 
              rdfService.GetLiteral(newName),
              true);
    ds.Assert(rNewType,
              rdfService.GetResource(Annozilla.NC + 'Icon'), 
              rdfService.GetLiteral(newIcon),
              true);
    ds.Assert(rNewType,
              rdfService.GetResource(Annozilla.NC + 'URL'), 
              rdfService.GetLiteral(newType),
              true);

    var RDFContainer = Components.classes['@mozilla.org/rdf/container;1']
        .getService(Components.interfaces.nsIRDFContainer);
    RDFContainer.Init(ds, rdfService.GetResource(Annozilla.ATYPES+'#annotationTypes'));
    RDFContainer.AppendElement(rNewType);
    
  } catch (e) {
      this.logToConsole (e);
      dump("  -- "+e+"\n");
  }
}

Annozilla.prototype.copyDataSource = function (dsource1, dsource2) {
  var sourcelist = dsource1.GetAllResources();
  while(sourcelist.hasMoreElements()){
    var source = sourcelist.getNext();
    var props = dsource1.ArcLabelsOut(source);
    while(props.hasMoreElements()){
      var prop = props.getNext();
      prop = prop.QueryInterface(Components.interfaces.nsIRDFResource);
      var target = dsource1.GetTarget(source,prop,true);
      if (target != null) dsource2.Assert(source,prop,target,true);
    }
  }
}
